@@ -24,9 +24,12 @@
  *  USA.
  */
 
+#define _GNU_SOURCE 1
+
 #include <string.h>
 #include <rfb/rfb.h>
 #include <rfb/rfbregion.h>
+#include <netdb.h>
 
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
@@ -61,31 +64,6 @@ rfbClientPtr pointerClient = NULL;  /* Mutex for pointer events */
 static void rfbProcessClientProtocolVersion(rfbClientPtr cl);
 static void rfbProcessClientNormalMessage(rfbClientPtr cl);
 
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-void rfbIncrClientRef(rfbClientPtr cl)
-{
-  LOCK(cl->refCountMutex);
-  cl->refCount++;
-  UNLOCK(cl->refCountMutex);
-}
-
-void rfbDecrClientRef(rfbClientPtr cl)
-{
-  LOCK(cl->refCountMutex);
-  cl->refCount--;
-  if(cl->refCount<=0) /* just to be sure also < 0 */
-    TSIGNAL(cl->deleteCond);
-  UNLOCK(cl->refCountMutex);
-}
-#else
-void rfbIncrClientRef(rfbClientPtr cl) {}
-void rfbDecrClientRef(rfbClientPtr cl) {}
-#endif
-
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-MUTEX(rfbClientListMutex);
-#endif
-
 struct rfbClientIterator {
   rfbClientPtr next;
   rfbScreenInfoPtr screen;
@@ -96,12 +74,11 @@ rfbClientListInit(rfbScreenInfoPtr rfbScreen)
 {
     if(sizeof(rfbBool)!=1) {
         /* a sanity check */
-        fprintf(stderr,"rfbBool's size is not 1 (%d)!\n",sizeof(rfbBool));
+        fprintf(stderr,"rfbBool's size is not 1 (%lu)!\n",(unsigned long)sizeof(rfbBool));
 	/* we cannot continue, because rfbBool is supposed to be char everywhere */
 	exit(1);
     }
     rfbScreen->rfbClientHead = NULL;
-    INIT_MUTEX(rfbClientListMutex);
 }
 
 rfbClientIteratorPtr
@@ -117,12 +94,6 @@ rfbGetClientIterator(rfbScreenInfoPtr rfbScreen)
 rfbClientPtr
 rfbClientIteratorHead(rfbClientIteratorPtr i)
 {
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-  if(i->next != 0) {
-    rfbDecrClientRef(i->next);
-    rfbIncrClientRef(i->screen->rfbClientHead);
-  }
-#endif
   LOCK(rfbClientListMutex);
   i->next = i->screen->rfbClientHead;
   UNLOCK(rfbClientListMutex);
@@ -137,38 +108,26 @@ rfbClientIteratorNext(rfbClientIteratorPtr i)
     i->next = i->screen->rfbClientHead;
     UNLOCK(rfbClientListMutex);
   } else {
-    IF_PTHREADS(rfbClientPtr cl = i->next);
     i->next = i->next->next;
-    IF_PTHREADS(rfbDecrClientRef(cl));
   }
 
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-    while(i->next && i->next->sock<0)
-      i->next = i->next->next;
-    if(i->next)
-      rfbIncrClientRef(i->next);
-#endif
-
     return i->next;
 }
 
 void
 rfbReleaseClientIterator(rfbClientIteratorPtr iterator)
 {
-  IF_PTHREADS(if(iterator->next) rfbDecrClientRef(iterator->next));
   free(iterator);
 }
 
-
 /*
  * rfbNewClientConnection is called from sockets.c when a new connection
  * comes in.
  */
 
 void
-rfbNewClientConnection(rfbScreen,sock)
-    rfbScreenInfoPtr rfbScreen;
-    int sock;
+rfbNewClientConnection(rfbScreenInfoPtr rfbScreen,
+                       int sock)
 {
     rfbClientPtr cl;
 
@@ -182,16 +141,17 @@ rfbNewClientConnection(rfbScreen,sock)
  */
 
 rfbClientPtr
-rfbNewClient(rfbScreen,sock)
-    rfbScreenInfoPtr rfbScreen;
-    int sock;
+rfbNewClient(rfbScreenInfoPtr rfbScreen,
+             int sock)
 {
     rfbProtocolVersionMsg pv;
     rfbClientIteratorPtr iterator;
     rfbClientPtr cl,cl_;
-    struct sockaddr_in addr;
-    size_t addrlen = sizeof(struct sockaddr_in);
+    struct sockaddr_storage addr;
+    socklen_t addrlen = sizeof(addr);
     int i;
+    char host[NI_MAXHOST];
+    const char *prt = "unknown";
 
     cl = (rfbClientPtr)calloc(sizeof(rfbClientRec),1);
 
@@ -205,7 +165,21 @@ rfbNewClient(rfbScreen,sock)
       int one=1;
 
       getpeername(sock, (struct sockaddr *)&addr, &addrlen);
-      cl->host = strdup(inet_ntoa(addr.sin_addr));
+
+      getnameinfo((struct sockaddr *)&addr,
+                  addrlen,
+                  host, sizeof(host),
+                  NULL, 0,
+                  0);
+
+      cl->host = strdup(host);
+
+      if (addr.ss_family == AF_INET)
+        prt = "IPv4";
+      else if (addr.ss_family == AF_INET6)
+        prt = "IPv6";
+
+      rfbLog("[%s] Got connection from client %s\n", prt, cl->host);
 
       rfbLog("  other clients:\n");
       iterator = rfbGetClientIterator(rfbScreen);
@@ -232,10 +206,6 @@ rfbNewClient(rfbScreen,sock)
       FD_SET(sock,&(rfbScreen->allFds));
 		rfbScreen->maxFd = max(sock,rfbScreen->maxFd);
 
-      INIT_MUTEX(cl->outputMutex);
-      INIT_MUTEX(cl->refCountMutex);
-      INIT_COND(cl->deleteCond);
-
       cl->state = RFB_PROTOCOL_VERSION;
 
       cl->readyForSetColourMapEntries = FALSE;
@@ -244,7 +214,7 @@ rfbNewClient(rfbScreen,sock)
       cl->correMaxWidth = 48;
       cl->correMaxHeight = 48;
 #ifdef HAVE_LIBZ
-      cl->zrleData = 0;
+      cl->zrleData = NULL;
 #endif
 
       cl->copyRegion = sraRgnCreate();
@@ -254,9 +224,6 @@ rfbNewClient(rfbScreen,sock)
       cl->modifiedRegion =
 	sraRgnCreateRect(0,0,rfbScreen->width,rfbScreen->height);
 
-      INIT_MUTEX(cl->updateMutex);
-      INIT_COND(cl->updateCond);
-
       cl->requestedRegion = sraRgnCreate();
 
       cl->format = cl->screen->rfbServerFormat;
@@ -265,7 +232,6 @@ rfbNewClient(rfbScreen,sock)
 
       LOCK(rfbClientListMutex);
 
-      IF_PTHREADS(cl->refCount = 0);
       cl->next = rfbScreen->rfbClientHead;
       cl->prev = NULL;
       if (rfbScreen->rfbClientHead)
@@ -303,10 +269,7 @@ rfbNewClient(rfbScreen,sock)
       if (WriteExact(cl, pv, sz_rfbProtocolVersionMsg) < 0) {
         rfbLogPerror("rfbNewClient: write");
         rfbCloseClient(cl);
-	/* TODO: memory leak here (cl is never freed)
-	 * can rfbClientConnectionGone called at this time?
-	 * tim@tjansen.de
-	 */
+        rfbClientConnectionGone(cl);
         return NULL;
       }
     }
@@ -338,8 +301,7 @@ rfbNewClient(rfbScreen,sock)
  */
 
 void
-rfbClientConnectionGone(cl)
-     rfbClientPtr cl;
+rfbClientConnectionGone(rfbClientPtr cl)
 {
     int i;
 
@@ -352,21 +314,13 @@ rfbClientConnectionGone(cl)
     if (cl->next)
         cl->next->prev = cl->prev;
 
+    if(cl->sock>0)
+        close(cl->sock);
+
 #ifdef HAVE_LIBZ
     FreeZrleData(cl);
 #endif
 
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-    if(cl->screen->backgroundLoop != FALSE)
-      do {
-	LOCK(cl->refCountMutex);
-	i=cl->refCount;
-	UNLOCK(cl->refCountMutex);
-	if(i>0)
-	  WAIT(cl->deleteCond,cl->refCountMutex);
-      } while(i>0);
-#endif
-
     if(cl->sock>=0)
        FD_CLR(cl->sock,&(cl->screen->allFds));
 
@@ -401,12 +355,6 @@ rfbClientConnectionGone(cl)
 
     if (cl->translateLookupTable) free(cl->translateLookupTable);
 
-    TINI_COND(cl->updateCond);
-    TINI_MUTEX(cl->updateMutex);
-
-    LOCK(cl->outputMutex);
-    TINI_MUTEX(cl->outputMutex);
-
     rfbPrintStats(cl);
 
     free(cl);
@@ -418,8 +366,7 @@ rfbClientConnectionGone(cl)
  */
 
 void
-rfbProcessClientMessage(cl)
-     rfbClientPtr cl;
+rfbProcessClientMessage(rfbClientPtr cl)
 {
     switch (cl->state) {
     case RFB_PROTOCOL_VERSION:
@@ -458,12 +405,10 @@ rfbProcessClientMessage(cl)
  */
 
 static void
-rfbProcessClientProtocolVersion(cl)
-    rfbClientPtr cl;
+rfbProcessClientProtocolVersion(rfbClientPtr cl)
 {
     rfbProtocolVersionMsg pv;
     int n, major_, minor_;
-    char failureReason[256];
 
     if ((n = ReadExact(cl, pv, sz_rfbProtocolVersionMsg)) <= 0) {
         if (n == 0)
@@ -477,7 +422,7 @@ rfbProcessClientProtocolVersion(cl)
     pv[sz_rfbProtocolVersionMsg] = 0;
     if (sscanf(pv,rfbProtocolVersionFormat,&major_,&minor_) != 2) {
         char name[1024]; 
-	if(sscanf(pv,"RFB %03d.%03d %1024s\n",&major_,&minor_,name) != 3) {
+	if(sscanf(pv,"RFB %03d.%03d %1023s\n",&major_,&minor_,name) != 3) {
 	    rfbErr("rfbProcessClientProtocolVersion: not a valid RFB client\n");
 	    rfbCloseClient(cl);
 	    return;
@@ -485,23 +430,26 @@ rfbProcessClientProtocolVersion(cl)
 	free(cl->host);
 	cl->host=strdup(name);
     }
-    rfbLog("Protocol version %d.%d\n", major_, minor_);
+    rfbLog("Client Protocol Version %d.%d\n", major_, minor_);
 
     if (major_ != rfbProtocolMajorVersion) {
-        /* Major version mismatch - send a ConnFailed message */
-        sprintf(failureReason,
-                "RFB protocol version mismatch - server %d.%d, client %d.%d",
-                rfbProtocolMajorVersion,rfbProtocolMinorVersion7,major_,minor_);
-        rfbClientConnFailed(cl, failureReason);
+        rfbErr("RFB protocol version mismatch - server %d.%d, client %d.%d",
+                rfbProtocolMajorVersion, rfbProtocolMinorVersion7,
+                major_,minor_);
+        rfbCloseClient(cl);
         return;
     }
 
-    if (minor_ != rfbProtocolMinorVersion3 && minor_ != rfbProtocolMinorVersion7) {
+    if (minor_ != rfbProtocolMinorVersion3 &&
+	minor_ != rfbProtocolMinorVersion7 &&
+	minor_ != rfbProtocolMinorVersion8) {
         /* Minor version mismatch - warn but try to continue */
         rfbLog("Ignoring minor version mismatch\n");
     }
 
-    if (minor_ >= rfbProtocolMinorVersion7)
+    if (minor_ >= rfbProtocolMinorVersion8)
+      cl->minorVersion = rfbProtocolMinorVersion8;
+    else if (minor_ == rfbProtocolMinorVersion7)
       cl->minorVersion = rfbProtocolMinorVersion7;
     else
       cl->minorVersion = rfbProtocolMinorVersion3;
@@ -516,19 +464,13 @@ rfbProcessClientProtocolVersion(cl)
  */
 
 void
-rfbClientConnFailed(cl, reason)
-    rfbClientPtr cl;
-    char *reason;
+rfbClientConnFailed(rfbClientPtr cl,
+                    char *reason)
 {
     char *buf;
     int len = strlen(reason);
 
-    rfbErr("Connection failed: %s\n", reason);
-
-    /* Note, for version 3.7 of the protocol, rfbConnFailed should
-     * be encoded in a single byte ... but we only send this message
-     * when we get an unknown major version so um ...
-     */
+    rfbLog("rfbClientConnFailed(\"%s\")\n", reason);
 
     buf = (char *)malloc(8 + len);
     ((uint32_t *)buf)[0] = 0;
@@ -538,6 +480,7 @@ rfbClientConnFailed(cl, reason)
     if (WriteExact(cl, buf, 8 + len) < 0)
         rfbLogPerror("rfbClientConnFailed: write");
     free(buf);
+
     rfbCloseClient(cl);
 }
 
@@ -548,8 +491,7 @@ rfbClientConnFailed(cl, reason)
  */
 
 void
-rfbProcessClientInitMessage(cl)
-    rfbClientPtr cl;
+rfbProcessClientInitMessage(rfbClientPtr cl)
 {
     rfbClientInitMsg ci;
     char buf[256];
@@ -567,6 +509,8 @@ rfbProcessClientInitMessage(cl)
         return;
     }
 
+    memset(buf,0,sizeof(buf));
+
     si->framebufferWidth = Swap16IfLE(cl->screen->width);
     si->framebufferHeight = Swap16IfLE(cl->screen->height);
     si->format = cl->screen->rfbServerFormat;
@@ -574,10 +518,7 @@ rfbProcessClientInitMessage(cl)
     si->format.greenMax = Swap16IfLE(si->format.greenMax);
     si->format.blueMax = Swap16IfLE(si->format.blueMax);
 
-    if (strlen(cl->screen->desktopName) > 128)      /* sanity check on desktop name len */
-        ((char*)cl->screen->desktopName)[128] = 0;
-
-    strcpy(buf + sz_rfbServerInitMsg, cl->screen->desktopName);
+    strncpy(buf + sz_rfbServerInitMsg, cl->screen->desktopName, 127);
     len = strlen(buf + sz_rfbServerInitMsg);
     si->nameLength = Swap32IfLE(len);
 
@@ -624,8 +565,7 @@ rfbProcessClientInitMessage(cl)
  */
 
 static void
-rfbProcessClientNormalMessage(cl)
-    rfbClientPtr cl;
+rfbProcessClientNormalMessage(rfbClientPtr cl)
 {
     int n=0;
     rfbClientToServerMsg msg;
@@ -723,51 +663,22 @@ rfbProcessClientNormalMessage(cl)
 		cl->useCopyRect = TRUE;
                 break;
             case rfbEncodingRaw:
-                if (cl->preferredEncoding == -1) {
-                    cl->preferredEncoding = enc;
-                    rfbLog("Using raw encoding for client %s\n",
-                           cl->host);
-                }
-                break;
             case rfbEncodingRRE:
-                if (cl->preferredEncoding == -1) {
-                    cl->preferredEncoding = enc;
-                    rfbLog("Using rre encoding for client %s\n",
-                           cl->host);
-                }
-                break;
             case rfbEncodingCoRRE:
-                if (cl->preferredEncoding == -1) {
-                    cl->preferredEncoding = enc;
-                    rfbLog("Using CoRRE encoding for client %s\n",
-                           cl->host);
-                }
-                break;
             case rfbEncodingHextile:
-                if (cl->preferredEncoding == -1) {
-                    cl->preferredEncoding = enc;
-                    rfbLog("Using hextile encoding for client %s\n",
-                           cl->host);
-                }
-                break;
 #ifdef HAVE_LIBZ
 	    case rfbEncodingZlib:
-		if (cl->preferredEncoding == -1) {
-		    cl->preferredEncoding = enc;
-		    rfbLog("Using zlib encoding for client %s\n",
-			   cl->host);
-		}
-              break;
+            case rfbEncodingZRLE:
 #ifdef HAVE_LIBJPEG
 	    case rfbEncodingTight:
-		if (cl->preferredEncoding == -1) {
-		    cl->preferredEncoding = enc;
-		    rfbLog("Using tight encoding for client %s\n",
-			   cl->host);
-		}
-		break;
 #endif
 #endif
+            /* The first supported encoding is the 'preferred' encoding */
+                if (cl->preferredEncoding == -1)
+                    cl->preferredEncoding = enc;
+
+
+                break;
 	    case rfbEncodingXCursor:
 		enableCursorShapeUpdates = TRUE;
 		break;
@@ -792,15 +703,6 @@ rfbProcessClientNormalMessage(cl)
 		    cl->useNewFBSize = TRUE;
 		}
 		break;
-#ifdef HAVE_LIBZ
-           case rfbEncodingZRLE:
-               if (cl->preferredEncoding == -1) {
-                   cl->preferredEncoding = enc;
-                   rfbLog("Using ZRLE encoding for client %s\n",
-                          cl->host);
-               }
-               break;
-#endif
             default:
 #ifdef HAVE_LIBZ
 		if ( enc >= (uint32_t)rfbEncodingCompressLevel0 &&
@@ -878,7 +780,6 @@ rfbProcessClientNormalMessage(cl)
 	    sraRgnOr(cl->modifiedRegion,tmpRegion);
 	    sraRgnSubtract(cl->copyRegion,tmpRegion);
        }
-       TSIGNAL(cl->updateCond);
        UNLOCK(cl->updateMutex);
 
        sraRgnDestroy(tmpRegion);
@@ -943,23 +844,23 @@ rfbProcessClientNormalMessage(cl)
 	    return;
 	}
 
+	msg.cct.length = Swap32IfLE(msg.cct.length);
+
+	str = (char *)malloc(msg.cct.length);
+
+	if ((n = ReadExact(cl, str, msg.cct.length)) <= 0) {
+	    if (n != 0)
+	        rfbLogPerror("rfbProcessClientNormalMessage: read");
+	    free(str);
+	    rfbCloseClient(cl);
+	    return;
+	}
+
 	if(!cl->viewOnly) {
-	    msg.cct.length = Swap32IfLE(msg.cct.length);
-
-	    str = (char *)malloc(msg.cct.length);
-
-	    if ((n = ReadExact(cl, str, msg.cct.length)) <= 0) {
-		if (n != 0)
-		    rfbLogPerror("rfbProcessClientNormalMessage: read");
-		free(str);
-		rfbCloseClient(cl);
-		return;
-	    }
-
 	    if (cl->screen->setXCutText)
 	      cl->screen->setXCutText(str, msg.cct.length, cl);
-	    free(str);
 	}
+	free(str);
 
         return;
 
@@ -983,11 +884,10 @@ rfbProcessClientNormalMessage(cl)
  */
 
 rfbBool
-rfbSendFramebufferUpdate(cl, givenUpdateRegion)
-     rfbClientPtr cl;
-     sraRegionPtr givenUpdateRegion;
+rfbSendFramebufferUpdate(rfbClientPtr cl,
+                         sraRegionPtr givenUpdateRegion)
 {
-    sraRectangleIterator* i = NULL;
+    sraRectangleIterator* i=NULL;
     sraRect rect;
     int nUpdateRegionRects;
     rfbFramebufferUpdateMsg *fu = (rfbFramebufferUpdateMsg *)cl->updateBuf;
@@ -1073,7 +973,7 @@ rfbSendFramebufferUpdate(cl, givenUpdateRegion)
     /*
      * The client is interested in the region requestedRegion.  The region
      * which should be updated now is the intersection of requestedRegion
-     * and the union of givenUpdateRegion and copyRegion.  If it's empty then
+     * and the union of modifiedRegion and copyRegion.  If it's empty then
      * no update is needed.
      */
 
@@ -1143,7 +1043,6 @@ rfbSendFramebufferUpdate(cl, givenUpdateRegion)
      * carry over a copyRegion for a future update.
      */
 
-
      sraRgnOr(cl->modifiedRegion,cl->copyRegion);
      sraRgnSubtract(cl->modifiedRegion,updateRegion);
      sraRgnSubtract(cl->modifiedRegion,updateCopyRegion);
@@ -1210,10 +1109,16 @@ rfbSendFramebufferUpdate(cl, givenUpdateRegion)
     fu->type = rfbFramebufferUpdate;
     if (nUpdateRegionRects != 0xFFFF) {
 	if(cl->screen->maxRectsPerUpdate>0
+	   /* CoRRE splits the screen into smaller squares */
+	   && cl->preferredEncoding != rfbEncodingCoRRE
+#ifdef HAVE_LIBZ
+	   /* Zlib encoding splits rectangles up into smaller chunks */
+	   && cl->preferredEncoding != rfbEncodingZlib
 #ifdef HAVE_LIBJPEG
 	   /* Tight encoding counts the rectangles differently */
 	   && cl->preferredEncoding != rfbEncodingTight
 #endif
+#endif /* HAVE_LIBZ */
 	   && nUpdateRegionRects>cl->screen->maxRectsPerUpdate) {
 	    sraRegion* newUpdateRegion = sraRgnBBox(updateRegion);
 	    sraRgnDestroy(updateRegion);
@@ -1331,10 +1236,10 @@ rfbSendFramebufferUpdate(cl, givenUpdateRegion)
  */
 
 rfbBool
-rfbSendCopyRegion(cl, reg, dx, dy)
-    rfbClientPtr cl;
-    sraRegionPtr reg;
-    int dx, dy;
+rfbSendCopyRegion(rfbClientPtr cl,
+                  sraRegionPtr reg,
+                  int dx,
+                  int dy)
 {
     int x, y, w, h;
     rfbFramebufferUpdateRectHeader rect;
@@ -1381,9 +1286,11 @@ rfbSendCopyRegion(cl, reg, dx, dy)
  */
 
 rfbBool
-rfbSendRectEncodingRaw(cl, x, y, w, h)
-    rfbClientPtr cl;
-    int x, y, w, h;
+rfbSendRectEncodingRaw(rfbClientPtr cl,
+                       int x,
+                       int y,
+                       int w,
+                       int h)
 {
     rfbFramebufferUpdateRectHeader rect;
     int nlines;
@@ -1454,8 +1361,7 @@ rfbSendRectEncodingRaw(cl, x, y, w, h)
  */
 
 rfbBool
-rfbSendLastRectMarker(cl)
-    rfbClientPtr cl;
+rfbSendLastRectMarker(rfbClientPtr cl)
 {
     rfbFramebufferUpdateRectHeader rect;
 
@@ -1486,9 +1392,9 @@ rfbSendLastRectMarker(cl)
  */
 
 rfbBool
-rfbSendNewFBSize(cl, w, h)
-    rfbClientPtr cl;
-    int w, h;
+rfbSendNewFBSize(rfbClientPtr cl,
+                 int w,
+                 int h)
 {
     rfbFramebufferUpdateRectHeader rect;
 
@@ -1520,8 +1426,7 @@ rfbSendNewFBSize(cl, w, h)
  */
 
 rfbBool
-rfbSendUpdateBuf(cl)
-    rfbClientPtr cl;
+rfbSendUpdateBuf(rfbClientPtr cl)
 {
     if(cl->sock<0)
       return FALSE;
@@ -1542,10 +1447,9 @@ rfbSendUpdateBuf(cl)
  */
 
 rfbBool
-rfbSendSetColourMapEntries(cl, firstColour, nColours)
-    rfbClientPtr cl;
-    int firstColour;
-    int nColours;
+rfbSendSetColourMapEntries(rfbClientPtr cl,
+                           int firstColour,
+                           int nColours)
 {
     char buf[sz_rfbSetColourMapEntriesMsg + 256 * 3 * 2];
     rfbSetColourMapEntriesMsg *scme = (rfbSetColourMapEntriesMsg *)buf;
