@@ -24,7 +24,6 @@
 
 #include "vino-util.h"
 #include <gtk/gtk.h>
-#include <string.h>
 
 #ifdef G_ENABLE_DEBUG
 VinoDebugFlags _vino_debug_flags = VINO_DEBUG_NONE;
@@ -42,7 +41,8 @@ vino_setup_debug_flags (void)
       { "tls",     VINO_DEBUG_TLS },
       { "mdns",    VINO_DEBUG_MDNS },
       { "prompt",  VINO_DEBUG_PROMPT },
-      { "http",    VINO_DEBUG_HTTP}
+      { "http",    VINO_DEBUG_HTTP },
+      { "dbus",    VINO_DEBUG_DBUS }
     };
   
   env_str = g_getenv ("VINO_SERVER_DEBUG");
@@ -102,112 +102,34 @@ vino_init_stock_items (void)
   initialized = TRUE;
 }
 
-char *
-vino_base64_unencode (const char *data)
+void
+vino_util_show_error (const gchar *title, const gchar *message, GtkWindow *parent)
 {
-  static const char *to_base64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-  const        char *p;
-  char              *retval;
-  int                i, length, count, rem;
-  int                qw, tw;
+  GtkWidget *d;
+  gchar     *t;
 
-  if (!data || (length = strlen (data)) <= 0)
-    return NULL;
+  if (title)
+    t = g_strdup (title);
+  else
+    t = g_strdup (_("An error has occurred:"));
 
-  p = data;
-  count = rem = 0;
-  while (length > 0)
-    {
-      int skip, vrfy, i;
+  d = gtk_message_dialog_new (parent,
+			      GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+			      GTK_MESSAGE_ERROR,
+			      GTK_BUTTONS_CLOSE,
+			      "%s",
+			      t);
+  g_free (t);
 
-      skip = strspn (p, to_base64);
+  if (message)
+    gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (d),
+					      "%s",
+					      message);
 
-      count  += skip;
-      p      += skip;
-      length -= skip;
-
-      if (length <= 0)
-	break;
-
-      vrfy = strcspn (p, to_base64);
-      
-      for (i = 0; i < vrfy; i++)
-	{
-	  if (g_ascii_isspace (p [i]))
-	    continue;
-
-	  if (p [i] != '=')
-	    return NULL;
-	  
-	  /* rem must be either 2 or 3, otherwise
-	   * no '=' should be here
-	   */
-	  if ((rem = count % 4) < 2)
-	    return NULL;
-
-	  /* end-of-message */
-	  break;
-	}
-
-      length -= vrfy;
-      p      += vrfy;
-    }
-
-  retval = g_new0 (char, (count / 4) * 3 + (rem ? rem - 1 : 0) + 1);
-
-  if (count <= 0)
-    return retval;
-
-  qw = tw = 0;
-  for (i = 0; data [i]; i++)
-    {
-      char c = data [i];
-      char bits;
-
-      if (g_ascii_isspace (c))
-	continue;
-
-      bits = 0;
-      if ((c >= 'A') && (c <= 'Z'))
-	{
-	  bits = c - 'A';
-	}
-      else if ((c >= 'a') && (c <= 'z'))
-	{
-	  bits = c - 'a' + 26;
-	}
-      else if ((c >= '0') && (c <= '9'))
-	{
-	  bits = c - '0' + 52;
-	}
-      else if (c == '=')
-	{
-	  break;
-	}
-
-      switch (qw++)
-	{
-	case 0:
-	  retval [tw + 0] = (bits << 2) & 0xfc;
-	  break;
-	case 1:
-	  retval [tw + 0] |= (bits >> 4) & 0x03;
-	  retval [tw + 1]  = (bits << 4) & 0xf0;
-	  break;
-	case 2:
-	  retval [tw + 1] |= (bits >> 2) & 0x0f;
-	  retval [tw + 2]  = (bits << 6) & 0xc0;
-	  break;
-	case 3:
-	  retval [tw + 2] |= bits & 0x3f;
-	  qw = 0;
-	  tw +=3;
-	  break;
-	default:
-	  g_assert_not_reached ();
-	  break;
-	}
-    }
-
-  return retval;
+  g_signal_connect_swapped (d,
+			    "response", 
+			    G_CALLBACK (gtk_widget_destroy),
+			    d);
+  gtk_widget_show_all (GTK_WIDGET(d));
 }
+
